require 'securerandom'
require 'active_support/core_ext'
require 'redis-objects'

require 'query_dam/version'
require 'query_dam/concerns/trackable'

module QueryDam

  DEFAULT_EXPIRATION = 5.minutes

  class << self
    def watch_relation(relation)
      tracking_key = SecureRandom.uuid
      query = relation.select("`#{relation.model.table_name}`.`id`").to_sql
      query_hash = Redis::HashKey.new(tracking_key)
      query_hash.bulk_set(model: relation.model.name, query: query)
      query_hash.expire(expiration)
      Redis::Set.new(relation.model.name) << tracking_key
      tracking_key
    end

    def get_updates(tracking_key)
      query_hash = Redis::HashKey.new(tracking_key)
      return nil unless query_hash.exists?
      query_hash.expire(expiration)

      model = query_hash[:model].constantize
      additions_set, updates_set, removals_set = query_sets(tracking_key)

      updated_ids = (additions_set.members + updates_set.members).uniq
      removed_ids = removals_set.members

      result = {
        model: model,
        updates: model.where(id: updated_ids),
        exclusions: removed_ids.map(&:to_i) }

      [additions_set, updates_set, removals_set].each(&:clear)

      result
    end

    def query_sets(tracking_key)
      %w[additions updates removals].map do |set_type|
        Redis::Set.new("#{tracking_key} #{set_type}")
      end
    end

    def expiration=(value)
      @expiration = value
    end

    def expiration
      @expiration || DEFAULT_EXPIRATION
    end
  end
end
